/*
 * cprot.c: handles special case PRIVMSGs and NOTICEs to enable a
 * client-to-client messaging protocol for getting information such as client
 * versions, etc from other clients automatically 
 *
 *
 * Written By Michael Sandrof <ms5n@andrew.cmu.edu> 
 *
 * Copyright(c) 1990 
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */
/*
 * The client-to-client protocol is performed by sending queries to a foreign
 * client, in the form of PRIVMSGs, and receiving replys from that client in
 * the form of NOTICEs.  The format of the queries is as follows: 
 *
 * PRIVMSG <nick> :^ACLIENT <cmd> <arg list>^A<Text Data>^A 
 *
 * The ^A is a control-A character (character 1).  The <cmd> may be one of the
 * following: 
 *
 * VERSION		request client name and version info 
 *
 * USERINFO 		request user-settable information 
 *
 * CLIENTINFO 		request info about clients implementation of
 * protocol.  Arg 1 may optionally be a <cmd> to get information on. 
 *
 * ERRMSG 		send and ERRMSG to a client 
 *
 * The <arg list> is an optional list of words which are included depending on
 * which command is being used.  In CLIENTINFO, for example, and optional
 * argument may be included to specify information on a particular command. 
 *
 * The <Text Data> field is an optional field which can included any information
 * depending on which command is specified.  In most cases, this is filled in
 * with Human readable text so that people with clients that don't understand
 * this protocol can read and understand whats going on. 
 *
 * The replies from a foreign client are formatted as follows: 
 *
 * NOTICE <nick> :^ACLIENT <cmd> <arg list>^A<Text Data>^A 
 *
 * <cmd> is the same as above, and each <cmd> request should use the same <cmd>
 * in the reply.  The exception to this in if an error condition occurs, then
 * ERRMSG is used. 
 *
 * The <arg list> and <Text Data> fields are dependant on which <cmd> is being
 * replied to.  Here is a list: 
 *
 * VERSION		Arg 1: client name; Arg 2: client major version; Arg 3:
 * client minor version; Arg 4: system type of client;Text Data: Anything you
 * want 
 *
 * USERINFO		No args;Text Data: Any user definable text 
 *
 * CLIENTINFO		Arg 1: optionally the <cmd> sent in the request; Text
 * Data: contains requested information 
 *
 * ERRMSG		Arg 1: offending word that caused error; Text Data: error
 * message 
 *
 * Here are some examples of request and reply pairs (in raw form as sent to
 * each respective server).  The PRIVMSGs are sent from Cheese to BigCheese,
 * the NOTICEs from BigCheese to Cheese. 
 *
 * PRIVMSG BigCheese :^ACLIENT VERSION^AChecking your clients version^A 
 *
 * NOTICE Cheese :^ACLIENT VERSION IRCII 2 03 Unix^AThe client that makes a
 * difference^A 
 *
 *
 * PRIMSG BigCheese :^ACLIENT Hey har hoooo^AThis is bogus^A 
 *
 * NOTICE Cheese :^ACLIENT ERRMSG Hey^AUnknown keyword^A 
 *
 *
 * PRIVMSG BigCheese :^ACLIENT CLIENTINFO^AChecking your clientinfo^A 
 *
 * NOTICE Cheese :^ACLIENT CLIENTINFO^ASupported commands: VERSION USERINFO
 * CLIENTINFO ERRMSG^A 
 *
 *
 * PRIVMSG BigCheese :^ACLIENT CLIENTINFO VERSION^AChecking your clientinfo^A 
 *
 * NOTICE Cheese :^ACLIENT CLIENTINFO VERSION^AReturns information about client
 * type and version^A 
 *
 */
#include "irc.h"
#include "ircaux.h"
#include "cprot.h"
#include "vars.h"

#define CDELIM '\001'

#define VERSION 0
#define USERINFO 1
#define CLIENTINFO 2
#define ERRMSG 3

static char *messages[] =
{"VERSION",
 "USERINFO",
 "CLIENTINFO",
 "ERRMSG"};

static char *next_field(line, ret)
char *line,
   **ret;
{
    char *ptr;

    if (line) {
	if (ptr = index(line, CDELIM))
	    *(ptr++) = null(char);
	*ret = ptr;
    }
    return (line);
}

/* clientinfo: handles CLIENTINFO requests from another client */
static void clientinfo(from, line)
char *from,
    *line;
{
    char *word;

    if (word = next_arg(line, &line)) {
	if (stricmp(word, messages[VERSION]) == 0)
	    reply_notice(from, messages[CLIENTINFO], "Returns the name and version of client being used", messages[VERSION]);
	else if (stricmp(word, messages[USERINFO]) == 0)
	    reply_notice(from, messages[CLIENTINFO], "Returns some user settable information", messages[USERINFO]);
	else if (stricmp(word, messages[CLIENTINFO]) == 0)
	    reply_notice(from, messages[CLIENTINFO], "Returns information about client commands.  You may send the name of a command as an argument to get info on that command", messages[CLIENTINFO]);
	else if (stricmp(from, messages[ERRMSG]) == 0)
	    reply_notice(from, messages[CLIENTINFO], "This is used to report error conditions back to the originating client.", messages[ERRMSG]);
	else
	    reply_notice(from, messages[ERRMSG], "Unknown CLIENTINFO command", "%s", word);
    } else
	reply_notice(from, messages[CLIENTINFO], "VERSION USERINFO CLIENTINFO ERRMSG", empty_string);
}

/* version: handles VERSION requests from another client */
static void version(from)
char *from;
{
    char *major = null(char *),
        *minor;

    malloc_strcpy(&major, irc_version);
    if (minor = index(major, '.'))
	*(minor++) = null(char);
    reply_notice(from, messages[VERSION], "The shape of things to come.", "IRCII %s %s UNIX", major, (minor ? minor : "0"));
    new_free(&major);
}

/*
 * client_msg_protocol: determine if the given line is in the proper
 * client-client protocol.  If not, 0 is returned and the called can do what
 * it wants.  If so, it is handled according to the client-to-client protocol 
 */
int client_msg_protocol(from, line)
char *line,
    *from;
{
    char *cmd,
        *args,
        *rest;

    if ((*line == CDELIM) && (*(line + strlen(line) - 1) == CDELIM)) {
	line++;
	if (cmd = next_field(line, &rest)) {
	    if ((cmd = next_arg(cmd, &args)) == null(char *))
		return (0);
	    if (stricmp(cmd, "CLIENT") != 0)
		return (1);
	    if (cmd = next_arg(args, &args)) {
		put_it("*** %s request from %s", cmd, from);
		if (stricmp(cmd, messages[VERSION]) == 0)
		    version(from);
		else if (stricmp(cmd, messages[USERINFO]) == 0) {
		    rest = get_string_var(USERINFO_VAR);
		    reply_notice(from, messages[USERINFO], (rest ? rest : empty_string), empty_string);
		} else if (stricmp(cmd, messages[ERRMSG]) == 0)
		    reply_notice(from, messages[ERRMSG], "No error", "%s", args);
		else if (stricmp(cmd, messages[CLIENTINFO]) == 0)
		    clientinfo(from, args);
		else
		    reply_notice(from, messages[ERRMSG], "Unknown keyword", "%s", cmd);
	    }
	}
	return (1);
    } else
	return (0);
}

/*
 * client_notice_protocol: handles replies from client-to-client protocol
 * requests.  If the line isn't in proper format, 0 is returned and the
 * calling processes can handle it as it wants.  If it is in the proper
 * format, the data in the reply is decoded and displayed nicely. 
 */
int client_notice_protocol(from, line)
char *from;
char *line;
{
    char *cmd,
        *args,
        *rest,
        *field;

    if ((*line == CDELIM) && (*(line + strlen(line) - 1) == CDELIM)) {
	line++;
	if (cmd = next_field(line, &rest)) {
	    if ((cmd = next_arg(cmd, &args)) == null(char *))
		return (0);
	    if (stricmp(cmd, "CLIENT") != 0)
		return (1);
	    if (cmd = next_arg(args, &args)) {
		if ((field = next_field(rest, &rest)) == null(char *))
		    field = empty_string;
		if (stricmp(cmd, messages[VERSION]) == 0)
		    put_it("*** Client info for %s: %s : %s", from, args, field);
		else if (stricmp(cmd, messages[USERINFO]) == 0)
		    put_it("*** User info for %s: %s", from, field);
		else if (stricmp(cmd, messages[CLIENTINFO]) == 0) {
		    if (cmd = next_arg(args, &args))
			put_it("*** Clientinfo for %s: %s: %s", from, cmd, field);
		    else
			put_it("*** Clientinfo for %s: %s", from, field);
		} else if (stricmp(cmd, messages[ERRMSG]) == 0)
		    put_it("*** Error message from %s: %s: %s", from, args, field);
	    }
	}
	return (1);
    } else
	return (0);
}

/*
 * reply_notice: sends a properly formatted reply NOTICE to the nickname
 * given.  The cmd parameter spcifies which of the client-to-client protocols
 * is being replies to.  The text parameter specifies the data portion of the
 * reply.  The format and variable arguments are used to specify, in string
 * format, the reply arguments. 
 */
static int reply_notice(to, cmd, text, format, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10)
char *to,
    *text,
    *cmd,
    *format;
int arg1,
    arg2,
    arg3,
    arg4,
    arg5,
    arg6,
    arg7,
    arg8,
    arg9,
    arg10;
{
    sprintf(buffer, format, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10);
    send_to_server("NOTICE %s :\1CLIENT %s %s\1%s\1", to, cmd, buffer, text);
    return (0);
}
